#!/usr/bin/env python


import re

from os.path import exists
from xml.sax import parse
from xml.sax.handler import ContentHandler

from lib.settings import ERRORS_XML
from lib.settings import MSSQL_XML


class htmlHandler(ContentHandler):
    """
    This class defines methods to parse the input HTML page to
    fingerprint the remote database management system

    @author: Bernardo Damele
    """

    def __init__(self, page):
        self.__dbms = None
        self.__page = page
        self.__regexp = None
        self.__match = None

        self.dbms = None


    def startElement(self, name, attrs):
        if name == "dbms":
            self.__dbms = attrs.get("value")

        if name == "error":
            self.__regexp = attrs.get("regexp")
            self.__match = re.search(self.__regexp, self.__page, re.I)

            if self.__match:
                self.dbms = self.__dbms
                self.__match = None


class bannerHandler(ContentHandler):
    """
    This class defines methods to parse and extract information from
    the given DBMS banner based upon the data in XML file

    @author: Bernardo Damele
    """

    def __init__(self, banner):
        self.__banner = banner
        self.__release = None
        self.__regexp = None
        self.__match = None

        self.release = None
        self.version = None
        self.servicepack = None


    def startElement(self, name, attrs):
        if name == "dbms":
            self.__release = attrs.get("release")
            self.__regexp = attrs.get("regexp")
            self.__match = re.search(self.__regexp, self.__banner)

        if name == "info" and self.__match:
            self.release = self.__release
            self.version = attrs.get("version")
            self.servicepack = attrs.get("servicepack")


def __sanitizeCheck(string, xmlfile):
    """
    This function sanitizes the input string and checks the
    existance of XML file
    """

    if not exists(xmlfile):
        raise Exception, "unable to parse file '%s'" % xmlfile

    string = string.replace("\n", "").replace("\r", "")

    return string


def bannerParser(banner, xmlfile):
    """
    This function calls a class to extract information from the given
    DBMS banner based upon the data in XML file
    """

    banner = __sanitizeCheck(banner, xmlfile)
    handler = bannerHandler(banner)
    parse(xmlfile, handler)

    return handler.release, handler.version, handler.servicepack


def htmlParser(page, xmlfile):
    """
    This function calls a class that parses the input HTML page to
    fingerprint the remote database management system
    """

    page = __sanitizeCheck(page, xmlfile)
    handler = htmlHandler(page)
    parse(xmlfile, handler)

    return handler.dbms
